/*
 * Copyright (c) 2009-2014 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <dzo@simtreas.ru>
 *
 * See REAME (rus)
 */

#include "libbb.h"
#include "arch_crc.h"
#include <fcntl.h>

struct ARCH_CRC_LIST {
	off_t offset;
	off_t filesize;
	struct ARCH_CRC_LIST *next;
};

char *archive;
#define MAX_24 16777216

#define CRC_BD ".crc_bd"

static FILE *fcrc;
static off_t offset_end;
static int dbfiles;     /* counter (debug) */

static struct ARCH_CRC_LIST arch_crc_list[MAX_24];

#include "crc24.h"
#define CRC24_INIT 0xB704CEL

#define NO_FILE ((off_t)(-1ll))

static void store_crc(off_t offset, off_t filesize, unsigned int crc)
{
	struct ARCH_CRC_LIST *l;

	l = &arch_crc_list[crc];
	if(l->offset != NO_FILE) {
		/* crc collision */
		struct ARCH_CRC_LIST *lm = xmalloc(sizeof(struct ARCH_CRC_LIST));

		lm->next = l->next;
		l->next = lm;
		l = lm;
	}
	l->offset = offset;
	l->filesize = filesize;
}

static char *read_record(size_t *p_record_size)
{
	static char *buf;
	static size_t bufsiz;
	int c;
	char *p;

	if(buf == NULL)
		buf = xmalloc(bufsiz = BUFSIZ);
	p = buf;
	while((c = getc(fcrc)) != EOF) {
		if((buf + bufsiz) == p) {
			buf = realloc(buf, bufsiz * 2);
			if(buf == NULL)
				bb_error_msg_and_die("Can not allocate memory for read very long record from archive crc data base");
			p = buf + bufsiz;       /* may be new ptr */
			bufsiz *= 2;
		}
		*p++ = c;
		if(c == 0)
			break;
	}
	if(bb_xferror(fcrc, "archive crc data base"))
		exit(67);
	if(p == buf)
		return NULL;    /* EOF */
	if(p_record_size != NULL)
		*p_record_size = (p-buf);
	return buf;
}

static char *parse_record(char *record, unsigned int *pcrc, off_t *pfilesize)
{
	char *e;

	*pcrc = safe_strtol16(record, &e);
	if(errno)
		bb_error_msg_and_die("the crc database structure corrupted, strange a crc value");
	if(*e++ != ' ')
		bb_error_msg_and_die("the crc database structure corrupted, strange a delimiter");
	record = e;
	*pfilesize = safe_strtoll16(record, &e);
	if(errno)
		bb_error_msg_and_die("the crc database structure corrupted, strange a file size value");
	if(*e++ != ' ')
		bb_error_msg_and_die("the crc database structure corrupted, strange a delimiter");
	return e; /* filename */
}

void init_crc_list(int fix_daemon)
{
	int i;
	size_t record_size;
	off_t offset = 0;
	char *record;

	for(i = 0; i < MAX_24; i++)
		arch_crc_list[i].offset = NO_FILE;
	record = concat_path_file(archive, CRC_BD);

	if(fix_daemon < 0) {
		fcrc = xwafopen(record, "w+\0crc");     /* create only */
	} else {
		fcrc = fopen(record, "r+");     /* read all and add */
		if(fcrc == NULL) {
			bb_perror_msg("can not open %s", record);
			bb_error_msg_and_die("use the fix crc base option first");
		}
		if(fix_daemon > 0)      /* debug */
			bb_error_msg("load crc database");
	}
	free(record);

	while((record = read_record(&record_size)) != NULL) {
		unsigned int crc;
		off_t filesize;
		char *filename = parse_record(record, &crc, &filesize);

		if(fix_daemon > 0)      /* debug */
			bb_error_msg("load %d-th file %s(%lld) from crc database", ++dbfiles, filename, (long long)filesize);
		store_crc(offset, filesize, crc);
		offset += record_size;
	}
	offset_end = offset;
}

static int write2crc_db(unsigned int crc, off_t fs, const char *name)
{
	int r = fprintf(fcrc, "%x %llx %s%c", crc, (unsigned long long)fs, name, 0);
	if(r < 0)
		bb_perror_msg_and_die("Error write to crc database");
	store_crc(offset_end, fs, crc);
	return r;
}

static char *mkdir_date(char *d1, int d, int type_idx, int free_d1)
{
	char s[8];
	char *new_name;
	static int tst_exist[3];        /* maked already */

	sprintf(s, type_idx == 0 ? "%d" : "%02d", d);
	new_name = concat_path_file(d1, s);
	if(tst_exist[type_idx] != d) {
		/* changed YYYY/MM/DD or first call */
		tst_exist[type_idx] = d;
		/* MM DD probable is not exist if changed YYYY or MM */
		while(++type_idx < 3)
			tst_exist[type_idx] = 0;
		if(mkdir(new_name, 0755) != 0 && errno != EEXIST)
			bb_perror_msg_and_die("mkdir(%s)", new_name);
	}
	if(free_d1)
		free(d1);
	return new_name;
}

char *make_archive_name(const char *name)
{
	time_t t;
	struct tm *ptm;
	char *out;
	char *new_name;
	char *s;
	int r;

	/* make archive/YYYY/MM/DD */
	time(&t);
	ptm = localtime (&t);
	out = mkdir_date(archive, ptm->tm_year + 1900, 0, 0);
	out = mkdir_date(out, ptm->tm_mon + 1, 1, 1);
	out = mkdir_date(out, ptm->tm_mday, 2, 1);

	/* make archive/YYYY/MM/DD/SUBDIR/name */
	new_name = concat_path_file(out, name);
	s = new_name + strlen(out); /* archive/YYYY/MM/DD/ */
	free(out);
	for(s++; (s = strchr(s, '/')) != NULL; s++) {
		/* name have subdirs too */
		*s = '\0';
		if(mkdir(new_name, 0755) != 0 && errno != EEXIST)
			bb_perror_msg_and_die("mkdir(%s)", new_name);
		*s = '/';
	}
	out = new_name;

	/* make numbered suffix for eq names, but may be different body */
	for(r = 1, new_name = out; access(new_name, F_OK) == 0; r++) {
		/* found exist name, make suffix .N */
		if(new_name != out)
			free(new_name);
		new_name = bb_xasprintf("%s.%d", out, r);
	}
	if(new_name != out) {
		free(out);
		out = new_name;
	}
	return out;
}

static int
cmp_map__arch_file(const char *arch_file, FILE *in, off_t fs)
{
	int c1, c2;
	off_t char_pos = 0;
	FILE *afp = xwafopen(arch_file, "r\0archive");

	rewind(in);

	do {
		c1 = getc(afp);
		c2 = getc(in);
		if (c1 != c2) {
			if (c2 == EOF) {
				fclose(afp);
				if(ferror(in))
					return STATUS_IOE;
				if(char_pos < fs)
					return STATUS_SIZE;     /* hmm, test file is shorted now */
				bb_error_msg_and_die("%s changed size in archive, added more than %lld",
						arch_file, (long long)fs);
			}
			if (c1 == EOF) {
				if(bb_xferror(afp, arch_file))
					exit(68);
				if(char_pos < fs)
					bb_error_msg_and_die("%s changed size in archive, shorted from %lld to %lld",
							arch_file, (long long)fs, (long long)char_pos);
				fclose(afp);
				return STATUS_SIZE;     /* hmm, test file is growed now */
			}
			fclose(afp);
			return NEW_UNIQ;
		}
		++char_pos;
	} while (c1 != EOF);

	if(bb_xferror(afp, arch_file))
		exit(68);
	fclose(afp);
	return STATUS_DUPA;
}

int find_from_db(FILE *in, off_t filesize, char **exist, unsigned int crc)
{
	if(arch_crc_list[crc].offset != NO_FILE) {
		struct ARCH_CRC_LIST *l;

		for(l = &arch_crc_list[crc]; l != NULL; l = l->next) {
			/* archive have file(s) with collision this crc */
			if(l->filesize == filesize) {
				/* filesize eq, test file eq */
				char *filename, *record;
				unsigned int crc_in;
				off_t fsz_in;
				int cmp_rez;

				fseeko(fcrc, l->offset, SEEK_SET);
				record = read_record(NULL);
				if(record == NULL)
					bb_perror_msg_and_die("the crc database corrupted, unexpected EOF");
				filename = parse_record(record, &crc_in, &fsz_in);

				/* paranoia check */
				if(crc_in != crc)
					bb_error_msg_and_die("the crc database corrupted, the crc is not compared");
				if(fsz_in != filesize)
					bb_error_msg_and_die("the crc database corrupted, the filesize is not compared");

				/* compare */
				cmp_rez = cmp_map__arch_file(filename, in, filesize);
				if(cmp_rez == STATUS_DUPA)
					*exist = filename;
				if(cmp_rez != NEW_UNIQ)
					return cmp_rez;
			}
		}
	}
	return NEW_UNIQ;
}

void store_crc_file(const char *name, off_t filesize, unsigned int crc)
{
	fseeko(fcrc, offset_end, SEEK_SET);
	offset_end += write2crc_db(crc, filesize, name);
	fflush(fcrc);
	if(!daemon_mode)
		bb_error_msg("save %d-th file %s to crc database", ++dbfiles, name);
}


char bufc[4096];

int read_file_crc_tst0_out(const char *in_name, FILE *in, unsigned int *pcrc, int nobinary,
				off_t *pfs, const char *out_name, const char *wafopen_mode)
{
	off_t fs = 0;
	unsigned int crc = CRC24_INIT;
	int ret = 0;
	FILE *out = out_name != NULL ? xwafopen(out_name, wafopen_mode) : NULL;

	while(ret == 0) {
		size_t rd;
		const char *o;

		if(feof(in))
			break;
		rd = fread (bufc, 1, sizeof(bufc), in);
		if(bb_xferror(in, in_name)) {
			ret = STATUS_IOE;
			break;
		}
		fs += rd;

		if(nobinary) {
			size_t wr;

			o = bufc;
			for(wr = 0; wr < rd; wr++) {
				if(U_C(*o) < ' ' && *o != '\r' && *o != '\n' &&
				   *o != '\t' && *o != '\v' && *o != '\f')
					break;
				o++;
			}
			if(wr < rd) {
				log_msg("%s is binary, skiped", in_name);
				ret = STATUS_BIN;
				break;
			}
		}

		if(pcrc != NULL) {
			size_t len = rd;
			o = bufc;
			while (len--)
				crc = ((crc << 8) ^ crc24Table[((crc >> 16) ^ (U_C(*o++))) & 0xff]);
		}
		if(out != NULL) {
			o = bufc;
			while(rd > 0) {
				size_t wr = fwrite (o, 1, rd, out);
				rd -= wr;
				o += wr;
				if(ferror(out))
					bb_perror_msg_and_die("output file problem for %s", out_name);
			}
		}
	}
	if(pcrc != NULL)
		*pcrc = crc & 0xFFFFFFL;
	if(*pfs != fs)
		ret = STATUS_SIZE;
	*pfs = fs;
	if(out != NULL)
		fclose(out);
	else
		fclose(in);
	return ret;
}

static void fix_add_file(const char *name, off_t sz)
{
	unsigned int crc;
	FILE *in = xwafopen(name, "r\0archive");

	if(read_file_crc_tst0_out(name, in, &crc, 0, &sz, NULL, NULL) == STATUS_IOE)
		exit(66);
	/* ignore STATUS_SIZE - we do not paranoid */
	bb_error_msg("Store %d-th file %s (%lld bytes) into crc database", ++dbfiles, name, (long long)sz);
	offset_end += write2crc_db(crc, sz, name);
}

struct DLIST_S {
	char *data;
	size_t len;
	struct DLIST_S *link;
};

static inline struct DLIST_S *
dlist_add_to(struct DLIST_S *old_head, char *new_item, size_t len)
{
	struct DLIST_S *new_head;

	new_head = xmalloc(sizeof(struct DLIST_S));
	new_head->data = new_item;
	new_head->len = len;
	new_head->link = old_head;

	return(new_head);
}

#include <dirent.h>

static void recursive_scan_dir(const char *p)
{
    struct DLIST_S *dirs;
    size_t m0;
    char *fp0;

    dirs = dlist_add_to(NULL, (char *)p, p[1] == '\0' ? 0 : strlen(p));
    fp0 = xmalloc(m0 = 4096);
    /* emulate recursion */
    do {
	struct DLIST_S *d_add = NULL;   /* found subdirs */
	do {
	    DIR *dir = opendir(dirs->data);

	    if (dir != NULL) {
		struct dirent *de;
		size_t dirlen = dirs->len;

		while ((de = readdir(dir)) != NULL) {
			char *f = de->d_name;
			struct stat st;
			size_t fe_sz;
			size_t fsz;

			if(*f == '.')
				continue;

			fe_sz = strlen(f) + 1;
			/* dir / file \0 */
			fsz = dirlen + 1 + fe_sz;
			if(fsz > m0) {
				free(fp0);
				m0 = fsz;
				fp0 = xmalloc(m0);
			}
			if(dirlen)
				memcpy(fp0, dirs->data, dirlen);
			fp0[dirlen] = '/';
			memcpy(fp0 + dirlen + 1, f, fe_sz);
			if(lstat(fp0, &st))
				continue;
			if(S_ISDIR(st.st_mode)) {
				d_add = dlist_add_to(d_add, memcpy(xmalloc(fsz), fp0, fsz), fsz - 1);
			} else {
				if(S_ISREG(st.st_mode))
					fix_add_file(fp0, st.st_size);
			}
		}
		closedir(dir);
	    } else {
		if(daemon_mode)
			bb_perror_msg_and_die("opendir(%s)", dirs->data);
		else
			bb_perror_msg("opendir(%s)", dirs->data);
	    }
	    if(dirs->data != (char *)p) {
		struct DLIST_S *d = dirs;

		free(dirs->data);
		dirs = dirs->link;
		free(d);
	    } else
		break;
	} while(dirs);
	dirs = d_add;
    } while(dirs);
    free(fp0);
}

void fix_crc_file(void)
{
	init_crc_list(-1);
	recursive_scan_dir(archive);
	fflush(fcrc);
}
