/*
match.c -- ksh-like extended pattern matching for the shell and filename
	   globbing.
     .
  -     
  '\'.

 ,   glob:
  -  ,     '[/].',  
     .
  -     '/'     *  ?
  -   [:XXX:],  XXX - alnum alpha   

*/

#include "km.h"


#include <ctype.h>
#include <string.h>

static int gmatch (const char *string, const char *se, const char *pattern,
		      const char *pe);
static const char * brackmatch (const char *p, unsigned char test);
static int extmatch (int xc, const char *s, const char *se, const char *p,
			    const char *pe);
static const char * patscan (const char *string, const char *end, int delim);

/* The result of FOLD is an `unsigned char' */
# define FOLD(c) ((unsigned char)c)

#ifndef STREQ
#define STREQN(a, b, n) ((a)[0] == (b)[0] && strncmp(a, b, n) == 0)
#endif

/* We don't use strcoll(3) for range comparisons in bracket expressions,
   even if we have it, since it can have unwanted side effects in locales
   other than POSIX or US.  For instance, in the de locale, [A-Z] matches
   all characters.  So, for ranges we use ASCII collation, and for
   collating symbol equivalence we use strcoll().  The casts to int are
   to handle tests that use unsigned chars. */

#define rangecmp(c1, c2)        ((int)(c1) - (int)(c2))

int
match (const char *pattern, const char *string)
{
  const char *se, *pe;

  if (string == 0 || pattern == 0)
    return FNM_NOMATCH;

  se = string + strlen (string);
  pe = pattern + strlen (pattern);

  return (gmatch (string, se, pattern, pe));
}

/* Match STRING against the filename pattern PATTERN, returning zero if
   it matches, FNM_NOMATCH if not.  */
static int
gmatch (const char *string, const char *se, const char *pattern, const char *pe)
{
  const char *p, *n;  /* pattern, string */
  char c;             /* current pattern character */
  char sc;            /* current string character */

  p = pattern;
  n = string;

  if (string == 0 || pattern == 0)
    return FNM_NOMATCH;

  while (p < pe)
    {
      c = *p++;
      c = FOLD (c);

      sc = n < se ? *n : '\0';

      /* extmatch () will handle recursively calling gmatch, so we can
	 just return what extmatch() returns. */
      if (*p == '(' &&
	  (c == '+' || c == '*' || c == '?' || c == '@' || c == '!')) /* ) */
	{
	  return (extmatch (c, n, se, p, pe));
	}

      switch (c)
	{
	case '?':               /* Match single character */
	  if (sc == '\0')
	    return FNM_NOMATCH;
	  break;

	case '*':               /* Match zero or more characters */
	  if (p == pe)
	    return 0;

	  /* Collapse multiple consecutive, `*' and `?', but make sure that
	     one character of the string is consumed for each `?'. */
	  for (c = *p++; (c == '?' || c == '*'); c = *p++)
	    {
	      if (c == '?')
		{
		  if (sc == '\0')
		    return FNM_NOMATCH;
		  /* One character of the string is consumed in matching
		     this ? wildcard, so *??? won't match if there are
		     fewer than three characters. */
		  n++;
		  sc = n < se ? *n : '\0';
		}

	      /* Handle ******(patlist) */
	      if (c == '*' && *p == '(')  /*)*/
		{
		  const char *newn;
		  /* We need to check whether or not the extended glob
		     pattern matches the remainder of the string.
		     If it does, we match the entire pattern. */
		  for (newn = n; newn < se; ++newn)
		    {
		      if (extmatch (c, newn, se, p, pe) == 0)
			return (0);
		    }
		  /* We didn't match the extended glob pattern, but
		     that's OK, since we can match 0 or more occurrences.
		     We need to skip the glob pattern and see if we
		     match the rest of the string. */
		  newn = patscan (p + 1, pe, 0);
		  p = newn;
		}
	      if (p == pe)
		break;
	    }

	  /* If we've hit the end of the pattern and the last character of
	     the pattern was handled by the loop above, we've succeeded.
	     Otherwise, we need to match that last character. */
	  if (p == pe && (c == '?' || c == '*'))
	    return (0);

	  /* General case, use recursion. */
	  {
	    unsigned char c1;

	    c1 = FOLD (c);
	    for (--p; n < se; ++n)
	      {
		/* If we're doing an extended glob match and the pattern is not
		   one of the extended glob patterns, we can check the first
		   character. */
		if (p[1] != '(' && /*)*/
		    strchr ("?*+@!", *p) == 0 && c != '[' && FOLD (*n) != c1)
		  continue;

		/* Otherwise, we just recurse. */
		if (gmatch (n, se, p, pe) == 0)
		  return (0);
	      }
	    return FNM_NOMATCH;
	  }

	case '[':
	  {
	    if (sc == '\0' || n == se)
	      return FNM_NOMATCH;

	    p = brackmatch (p, sc);
	    if (p == 0)
	      return FNM_NOMATCH;
	  }
	  break;

	default:
	  if ((unsigned char)c != FOLD (sc))
	    return (FNM_NOMATCH);
	}

      ++n;
    }

  if (n == se)
    return (0);

  return (FNM_NOMATCH);
}

static const char *
brackmatch (const char *p, unsigned char test)
{
  char cstart, cend, c;
  int not;    /* Nonzero if the sense of the character class is inverted.  */
  int brcnt;
  const char *savep;

  test = FOLD (test);

  savep = p;

  /* POSIX.2 3.13.1 says that an exclamation mark (`!') shall replace the
     circumflex (`^') in its role in a `nonmatching list'.  A bracket
     expression starging with an unquoted circumflex character produces
     unspecified results.  This implementation treats the two identically. */
  if ((not = (*p == '!' || *p == '^')))
    ++p;

  c = *p++;
  for (;;)
    {
      /* Initialize cstart and cend in case `-' is the last
	 character of the pattern. */
      cstart = cend = c;

      cstart = cend = FOLD (cstart);

      /* POSIX.2 2.8.3.1.2 says: `An expression containing a `[' that
	 is not preceded by a backslash and is not part of a bracket
	 expression produces undefined results.'  This implementation
	 treats the `[' as just a character to be matched if there is
	 not a closing `]'. */
      if (c == '\0')
	return ((test == '[') ? savep : (char *)0);

      c = *p++;
      c = FOLD (c);

      /* This introduces a range, unless the `-' is the last
	 character of the class.  Find the end of the range
	 and move past it. */
      if (c == '-' && *p != ']')
	{
	  cend = *p++;
	  if (cend == '\0')
	    return (char *)0;
	  cend = FOLD (cend);

	  c = *p++;

	  /* POSIX.2 2.8.3.2:  ``The ending range point shall collate
	     equal to or higher than the starting range point; otherwise
	     the expression shall be treated as invalid.''  Note that this
	     applies to only the range expression; the rest of the bracket
	     expression is still checked for matches. */
	  if (rangecmp (cstart, cend) > 0)
	    {
	      if (c == ']')
		break;
	      c = FOLD (c);
	      continue;
	    }
	}

      if (rangecmp (test, cstart) >= 0 && rangecmp (test, cend) <= 0)
	goto matched;

      if (c == ']')
	break;
    }
  /* No match. */
  return (!not ? (char *)0 : p);

matched:
  /* Skip the rest of the [...] that already matched.  */
#if 0
  brcnt = (c != ']') + (c == '[' && (*p == '=' || *p == ':' || *p == '.'));
#else
  c = *--p;
  brcnt = 1;
#endif
  while (brcnt > 0)
    {
      /* A `[' without a matching `]' is just another character to match. */
      if (c == '\0')
	return ((test == '[') ? savep : (char *)0);

      c = *p++;
      if (c == '[' && (*p == '=' || *p == ':' || *p == '.'))
	brcnt++;
      else if (c == ']')
	brcnt--;
    }
  return (not ? (char *)0 : p);
}

/* ksh-like extended pattern matching:

	[?*+@!](pat-list)

   where pat-list is a list of one or patterns separated by `|'.  Operation
   is as follows:

	?(patlist)      match zero or one of the given patterns
	*(patlist)      match zero or more of the given patterns
	+(patlist)      match one or more of the given patterns
	@(patlist)      match exactly one of the given patterns
	!(patlist)      match anything except one of the given patterns
*/

/* Scan a pattern starting at STRING and ending at END, keeping track of
   embedded () and [].  If DELIM is 0, we scan until a matching `)'
   because we're scanning a `patlist'.  Otherwise, we scan until we see
   DELIM.  In all cases, we never scan past END.  The return value is the
   first character after the matching DELIM. */
static const char *
patscan (const char *string, const char *end, int delim)
{
  int pnest, bnest;
  const char *s;
  char c;

  pnest = bnest = 0;
  for (s = string; (c = *s); s++)
    {
      if (s >= end)
	return (s);
      switch (c)
	{
	case '\0':
	  return ((char *)0);
	case '[':
	  bnest++;
	  break;
	case ']':
	  if (bnest)
	    bnest--;
	  break;
	case '(':
	  if (bnest == 0)
	    pnest++;
	  break;
	case ')':
#if 0
	  if (bnest == 0)
	    pnest--;
	  if (pnest <= 0)
	    return ++s;
#else
	  if (bnest == 0 && pnest-- <= 0)
	    return ++s;
#endif
	  break;
	case '|':
	  if (bnest == 0 && pnest == 0 && delim == '|')
	    return ++s;
	  break;
	}
    }

  return (char *)0;
}

/* Return 0 if dequoted pattern matches S. */
static int
strcompare (const char *p, const char *pe, const char *s, const char *se)
{
  if((pe-p)!=(se-s))
    return FNM_NOMATCH;

  return (strncmp(p, s, (se-s)) == 0 ? 0 : FNM_NOMATCH);
}

/* Match a ksh extended pattern specifier.  Return FNM_NOMATCH on failure or
   0 on success.  This is handed the entire rest of the pattern and string
   the first time an extended pattern specifier is encountered, so it calls
   gmatch recursively. */
   /* xc select which operation */
static int
extmatch (int xc, const char *s, const char *se, const char *p, const char *pe)
{
  const char *prest;                  /* pointer to rest of pattern */
  const char *psub;                   /* pointer to sub-pattern */
  const char *pnext;                  /* pointer to next sub-pattern */
  const char *srest;                  /* pointer to rest of string */
  int m1, m2;

#if 0
fprintf(stderr, "extmatch: xc = %c\n", xc);
fprintf(stderr, "extmatch: s = %s; se = %s\n", s, se);
fprintf(stderr, "extmatch: p = %s; pe = %s\n", p, pe);
#endif

  prest = patscan (p + (*p == '('), pe, 0); /* ) */
  if (prest == 0)
    /* If PREST is 0, we failed to scan a valid pattern.  In this
       case, we just want to compare the two as strings. */
    return (strcompare (p - 1, pe, s, se));

  switch (xc)
    {
    case '+':                   /* match one or more occurrences */
    case '*':                   /* match zero or more occurrences */
      /* If we can get away with no matches, don't even bother.  Just
	 call gmatch on the rest of the pattern and return success if
	 it succeeds. */
      if (xc == '*' && (gmatch (s, se, prest, pe) == 0))
	return 0;

      /* OK, we have to do this the hard way.  First, we make sure one of
	 the subpatterns matches, then we try to match the rest of the
	 string. */
      for (psub = p + 1; ; psub = pnext)
	{
	  pnext = patscan (psub, pe, '|');
	  for (srest = s; srest <= se; srest++)
	    {
	      /* Match this substring (S -> SREST) against this
		 subpattern (psub -> pnext - 1) */
	      m1 = gmatch (s, srest, psub, pnext - 1) == 0;
	      /* OK, we matched a subpattern, so make sure the rest of the
		 string matches the rest of the pattern.  Also handle
		 multiple matches of the pattern. */
	      if (m1)
		m2 = (gmatch (srest, se, prest, pe) == 0) ||
		      (s != srest && gmatch (srest, se, p - 1, pe) == 0);
	      if (m1 && m2)
		return (0);
	    }
	  if (pnext == prest)
	    break;
	}
      return (FNM_NOMATCH);

    case '?':           /* match zero or one of the patterns */
    case '@':           /* match exactly one of the patterns */
      /* If we can get away with no matches, don't even bother.  Just
	 call gmatch on the rest of the pattern and return success if
	 it succeeds. */
      if (xc == '?' && (gmatch (s, se, prest, pe) == 0))
	return 0;

      /* OK, we have to do this the hard way.  First, we see if one of
	 the subpatterns matches, then, if it does, we try to match the
	 rest of the string. */
      for (psub = p + 1; ; psub = pnext)
	{
	  pnext = patscan (psub, pe, '|');
	  srest = (prest == pe) ? se : s;
	  for ( ; srest <= se; srest++)
	    {
	      if (gmatch (s, srest, psub, pnext - 1) == 0 &&
		  gmatch (srest, se, prest, pe) == 0)
		return (0);
	    }
	  if (pnext == prest)
	    break;
	}
      return (FNM_NOMATCH);

    case '!':           /* match anything *except* one of the patterns */
      for (srest = s; srest <= se; srest++)
	{
	  m1 = 0;
	  for (psub = p + 1; ; psub = pnext)
	    {
	      pnext = patscan (psub, pe, '|');
	      /* If one of the patterns matches, just bail immediately. */
	      if ((m1 = (gmatch (s, srest, psub, pnext - 1) == 0)))
		break;
	      if (pnext == prest)
		break;
	    }
	  if (m1 == 0 && gmatch (srest, se, prest, pe) == 0)
	    return (0);
	}
      return (FNM_NOMATCH);
    }

  return (FNM_NOMATCH);
}

#ifdef DEBUG
#include <stdio.h>

static void remove_cr(char *s)
{
	char *p;

	p = strchr(s, '\r');
	if(p != NULL)
		*p = '\0';
	p = strchr(s, '\n');
	if(p != NULL)
		*p = '\0';
}

int main(void)
{
	char patb[512], tst[512];

	printf("pattern: ");
	if(fgets(patb, sizeof(patb), stdin) == NULL)
		return 1;
	printf("Test (EOF for exit) pattern %s", patb);
	remove_cr(patb);
	for(;;) {
		printf("=> ");
		if(fgets(tst, sizeof(tst), stdin) == NULL)
			return 0;
		remove_cr(tst);
		puts(match(patb, tst) == FNM_NOMATCH ? "nomatch" : "match");
	}
}
#endif
