/*
 * Copyright (C) 2005-2012 by Vladimir Oleynik <dzo@simtreas.ru>
 * may safely be consumed by a GPL license.
 *
 * bb_simplify_path() - codepoet code:
 *      Copyright (C) 2005 Manuel Novoa III <mjn3@codepoet.org>
 *      my some bugreports
 *
 * bb_*error*():
 *      Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *      Hacked by me for use syslog in daemon mode
 *
 * xmalloc() bb_xstrdup():
 *      Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * last_char_is():
 *      Copyright (C) 2001 Larry Doolittle, <ldoolitt@recycle.lbl.gov>
 *
 * xgetcwd():
 *      Written by David MacKenzie <djm@gnu.ai.mit.edu>
 *      and me
 *
 * bb_xgetpwnam():
 *      Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *      Hacked by Tito Ragusa (c) 2004 <farmatito@tiscali.it>
 *      Hacked by me for integer/symbolic
 *
 * xwafopen():
 *      Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *      by me too - use type of name after \0 in mode string
 *
 * other functions by me
 */



#include "libbb.h"
#include <syslog.h>

/* external valiables */
const char *daemon_name;
int daemon_mode;
FILE *log_f;

static void bb_verror_msg(const char *s, va_list p)
{
	if(!daemon_mode) {
		fprintf(stderr, "%s: ", daemon_name);
		vfprintf(stderr, s, p);
	} else {
		if(log_f != NULL)
			vfprintf(log_f, s, p);
		else
			if(*s != '\0')
				vsyslog(LOG_ERR, s, p);
	}
}

void bb_perror_msg_and_die(const char *s, ...)
{
	int err = errno;
	va_list p;

	va_start(p, s);
	if(s == 0) s = "";
	bb_verror_msg(s, p);
	if (*s) s = ": ";

	if(!daemon_mode) {
		fprintf(stderr, "%s%s\n", s, strerror(err));
	} else {
		if(log_f != NULL)
			fprintf(log_f, "%s%s\n", s, strerror(err));
		else
			syslog(LOG_ERR, "Previous error: %s", strerror(err));
	}
	va_end(p);
	unlink_pidfile();
	exit(2);
}

void bb_perror_msg(const char *s, ...)
{
	int err = errno;
	va_list p;

	va_start(p, s);
	if(s == 0) s = "";
	bb_verror_msg(s, p);
	if (*s) s = ": ";

	if(!daemon_mode) {
		fprintf(stderr, "%s%s\n", s, strerror(err));
	} else {
		if(log_f != NULL) {
			fprintf(log_f, "%s%s\n", s, strerror(err));
			fflush(log_f);
		} else {
			syslog(LOG_ERR, "Previous error: %s", strerror(err));
		}
	}
	va_end(p);
}

void bb_error_msg_and_die(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	bb_verror_msg(s, p);
	if(!daemon_mode)
		putc('\n', stderr);
	else if(log_f != NULL)
		putc('\n', log_f);
	va_end(p);
	unlink_pidfile();
	exit(1);
}

void bb_error_msg(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	bb_verror_msg(s, p);
	if(!daemon_mode)
		putc('\n', stderr);
	else if(log_f != NULL) {
		putc('\n', log_f);
		fflush(log_f);
	}
	va_end(p);
}


void log_msg(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	if(!daemon_mode) {
		bb_verror_msg(s, p);
		putc('\n', stderr);
	} else if(log_f != NULL) {
		time_t ct;

		time(&ct);
		fprintf(log_f, "%.20s: ", ctime(&ct)+4);
		vfprintf(log_f, s, p);
		putc('\n', log_f);
		fflush(log_f);
	}
	va_end(p);
}

static char * last_char_is(const char *s, int c)
{
	char *sret = (char *)s;
	if (sret) {
		sret = strrchr(sret, c);
		if(sret != NULL && *(sret+1) != 0)
			sret = NULL;
	}
	return sret;
}

/* by me */
char *bb_xasprintf(const char *format, ...)
{
	va_list p;
	int r;
	char *string_ptr;

	va_start(p, format);
	r = vasprintf(&string_ptr, format, p);
	va_end(p);

	if (r < 0)
		bb_perror_msg_and_die("bb_xasprintf");
	return string_ptr;
}

/* by me bugreports */
char *concat_path_file(const char *path, const char *filename)
{
	char *lc;

	if (!path)
		path = "";
	lc = last_char_is(path, '/');
	while (*filename == '/')
		filename++;
	return bb_xasprintf("%s%s%s", path, (lc==NULL ? "/" : ""), filename);
}


static const char bb_msg_memory_exhausted[]="memory exhausted";

/* by me */
void free0 (char **p)
{
  if (*p != NULL) {
	free (*p);
	*p = NULL;
  }
}


void *xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if (ptr == NULL && size != 0)
		bb_error_msg_and_die(bb_msg_memory_exhausted);
	return ptr;
}

/* by me too */
static char *
xgetcwd (char *cwd)
{
  static const int PATH_INCR = 32;
  char *ret;
  unsigned path_max = 1026-PATH_INCR;

  if(cwd!=NULL)
	return cwd;

  do {
      path_max += PATH_INCR;
      free(cwd);
      cwd = xmalloc (path_max);
  } while ((ret = getcwd (cwd, path_max)) == NULL && errno == ERANGE);

  if (ret == NULL)
      bb_perror_msg_and_die("getcwd()");

  return cwd;
}

char * bb_xstrdup (const char *s) {
	char *t;

	if (s == NULL)
		return NULL;

	t = strdup (s);

	if (t == NULL)
		bb_error_msg_and_die(bb_msg_memory_exhausted);

	return t;
}


/* hack by me for const chdir and some bugreports */
char *bb_simplify_path(const char *path)
{
	char *s, *start, *p;
	static char *cwd;

	if (path[0] == '/')
		start = bb_xstrdup(path);
	else {
		cwd = xgetcwd(cwd);
		start = concat_path_file(cwd, path);
	}
	p = s = start;

	do {
		if (*p == '/') {
			if (*s == '/') {        /* skip duplicate (or initial) slash */
				continue;
			} else if (*s == '.') {
				if (s[1] == '/' || s[1] == 0) { /* remove extra '.' */
					continue;
				} else if ((s[1] == '.') && (s[2] == '/' || s[2] == 0)) {
					++s;
					if (p > start) {
						while (*--p != '/');    /* omit previous dir */
					}
					continue;
				}
			}
		}
		*++p = *s;
	} while (*++s);

	if ((p == start) || (*p != '/')) {      /* not a trailing slash */
		++p;                                    /* so keep last character */
	}
	*p = 0;

	return start;
}

/* by me */
long safe_strtol(const char *arg)
{
	char *endptr;
	long value = value;     /* make gcc happy */

	if(arg == NULL || *arg == '\0')
		goto eret;

	errno = 0;
	value = strtol(arg, &endptr, 0);
	if (*endptr != '\0') {
eret:
		errno = EINVAL;
	}
	return value;
}

/* by me */
long long safe_strtoll16(const char *arg, char **endptr)
{
	long long value = value;     /* make gcc happy */

	if(arg == NULL || *arg == '\0')
		goto eret;

	errno = 0;
	value = strtoll(arg, endptr, 16);
	if (**endptr != ' ') {
eret:
		errno = EINVAL;
	}
	return value;
}

/* by me */
unsigned int safe_strtol16(const char *arg, char **e)
{
	long value;
	unsigned int u32_ret;

	if(arg == NULL || *arg == '\0')
		goto eret;

	errno = 0;
	value = strtol(arg, e, 16);
	if (**e != ' ') {
eret:
		errno = EINVAL;
		return 0;
	}
	u32_ret = value;
	if((long)u32_ret != value)
		goto eret;
	return u32_ret;
}

#include <pwd.h>

/* by me too */
/* returns a uid given a username */
long bb_xgetpwnam(const char *name)
{
	struct passwd *myuser;
	long uid = safe_strtol(name);

	if(errno == 0) /* integer */
		return uid;

	myuser  = getpwnam(name);
	if (myuser == NULL)
		bb_error_msg_and_die("unknown user name: %s", name);

	return myuser->pw_uid;
}

/* by me */
static sigset_t bb_sigmask(unsigned int sigs_bits)
{
	unsigned int b = 2;
	int signo = 1;
	sigset_t m;

	sigemptyset(&m);
	while(b) {
		if((sigs_bits & b) != 0)
			sigaddset(&m, signo);
		signo++;
		b <<= 1;
	}
	return m;
}

/* by me */
sigset_t bb_sigblock(unsigned int sigs_bits)
{
	sigset_t m;

	m = bb_sigmask(sigs_bits);
	sigprocmask(SIG_BLOCK, &m, NULL);
	return m;
}

/* by me */
void bb_sigunblock(sigset_t m)
{
	sigprocmask(SIG_UNBLOCK, &m, NULL);
}

/* by me too */
extern const char usage_msg[];

void bb_show_usage (void)
{
	fprintf(stderr, "Usage %s %s\n", daemon_name, usage_msg);
	exit(3);
}

/* by me codepoet code, but simplified special for this program */
/* simplified:
   1.   GNU : "+" "-" "::"
   2.         
'int',          
  .
   3. applet_opts    "ab:..."   
\0,       \0 :
   -  '-'  ,         
 ,   
   -  ,     
:

	int flags;

	char *b_str = default_b_opt_value;
	flags = xgetopt(argc, argv, "ab:\0", &b_str);
        -a -b b_value  -


	char *b_str;
	flags = xgetopt(argc, argv, "ab:\0b", &b_str);
        -a   
 -b b_value  .     
b_str .

	char *b_str = default_b_opt_value;
	flags = xgetopt(argc, argv, "ab:\0-", &b_str);
	      -a -b b_value

	char *c_str;
	flags = xgetopt(argc, argv, "ab:c:\0-c", &b_str, &c_str);
     ,       -a 
-b b_value   -c c_value     

	 flags:
	if((flags & 1)) {      -a }
	if((flags & 2)) {      -b b_value }
	if((flags & 4)) {      -c c_value }
*/
#include <alloca.h>
int xgetopt(int argc, char **argv, const char *applet_opts, ...)
{
	const char *p;
	const char ***a;
	int c;
	int opt;
	int ret = 0;
	va_list vl;

	p = strrchr(argv[0], '/');
	if(p != NULL)
		daemon_name = p + 1;
	else
		daemon_name = argv[0];

	/* need applet_opts != NULL and applet_opts[0] != '\0' */
	a = alloca(strlen(applet_opts) * sizeof(char **));

	va_start (vl, applet_opts);
	for(c = 0, p = applet_opts; *p != '\0'; c++) {
		if(*++p == ':') {
			a[c] = va_arg (vl, const char **);
			p++;
		} else
			a[c] = NULL;
	}
	va_end(vl);

	while((opt=getopt(argc, argv, applet_opts)) != -1) {
		for(c = 0, p = applet_opts; *p != '\0'; c++) {
			if(*p == opt) {
				if(a[c] != NULL)
					*(a[c]) = optarg;
				if((unsigned)c >= (sizeof(int)*8))
					c = sizeof(int)*8 - 1;
				ret |= (1 << c);
				break;
			}
			if(*++p == ':')
				p++;
		}
		if(*p == '\0')
			bb_show_usage ();
	}

	argv += optind;
	/* test need options */
	for(p = applet_opts; *p != '\0'; p++)
		;
	if(*++p == '\0')
		return ret;     /* do not check need args */
	if(*p == '-') {
		p++;
		if(*argv == NULL)
			bb_show_usage ();       /* need arguments */
	} else if(*argv != NULL)                /* do not need arguments */
		bb_show_usage ();

	{
		const char *o = p;
		while(*o != '\0') {
			for(c = 0, p = applet_opts; *p != '\0'; c++) {
				if(p[1] == ':') {
					if(*o == *p && *a[c] == NULL)
						bb_show_usage ();
					p++;
				}
				p++;
			}
			o++;
		}
	}
	return ret;
}

/* by me too - use type of name after \0 */
FILE *xwafopen(const char *name, const char *xtype)
{
	FILE *out;
	const char *create_read_msg = (*xtype == 'w' || *xtype == 'a') ?
			"create" : "open for read";

	out = fopen(name, xtype);
	while(*xtype != '\0')
		xtype++;
	xtype++;
	if(out == NULL)
		bb_perror_msg_and_die("can not %s %s file %s",
					create_read_msg, xtype, name);
	return out;
}

int bb_xferror(FILE *fp, const char *fn)
{
	int r = ferror(fp);

	if (r)
		bb_error_msg("input error while read %s", fn);
	return r;
}

/* my function, no bb library */

void xstat(const char *name, struct stat *pst)
{
	if(stat(name, pst))
		bb_perror_msg_and_die("can not stat %s", name);
}

char *make_fullpath_and_tstisdir(const char *d)
{
	struct stat st;
	char *r = bb_simplify_path(d);

	xstat(d, &st);
	if(S_ISDIR(st.st_mode) == 0)
		bb_error_msg_and_die("%s is not directory", d);
	return r;
}

void tst_is_one_fs(const char *d1, const char *d2)
{
	struct stat st1, st2;

	xstat(d1, &st1);
	xstat(d2, &st2);
	if(st1.st_dev != st2.st_dev)
		bb_error_msg_and_die("%s and %s have different file system", d1, d2);
}

int tst_is_not_one_fs(const char *d1, const char *d2)
{
	struct stat st1, st2;

	xstat(d1, &st1);
	xstat(d2, &st2);
	if(st1.st_dev == st2.st_dev) {
		bb_error_msg("error: %s and %s have equivalent file system", d1, d2);
		return 1;
	}
	return 0;
}

#include <sys/file.h>

static const char *pidfilename;

void unlink_pidfile(void)
{
	if(pidfilename != NULL)
		unlink(pidfilename);
}

static void sigterm_handler(int sig)
{
	unlink_pidfile();
	exit(sig);
}

FILE *create_pid_file_and_lock(const char *pidfile)
{
	FILE *fl;

	/* create_pidfile */
	if(pidfile == NULL) {
		char *p = make_fullpath_and_tstisdir("/var/run");
		free(p);
		pidfile = bb_xasprintf("/var/run/%s.pid", daemon_name);
	}
	pidfilename = pidfile;

	fl = xwafopen(pidfile, "a\0pid");
	signal(SIGTERM, sigterm_handler);
	if(flock (fileno (fl), LOCK_EX | LOCK_NB) < 0) {
		pidfilename = NULL;
		bb_perror_msg_and_die("Detect already started or can not lock file in this filesystem");
	}
	return fl;
}

void write_pid(FILE *fl)
{
	ftruncate(fileno (fl), 0);
	fprintf(fl, "%d\n", getpid());
	fflush(fl);
}

void xdaemon(int change_dir_to_root, const char *pidfile)
{
	FILE *fl = create_pid_file_and_lock(pidfile);

	if(daemon(1, change_dir_to_root))       /* do not change dir if need */
		bb_perror_msg_and_die("daemon()");

	daemon_mode = 1;        /* stderr -> syslog and do not debug */
	write_pid(fl);
}
