/*
 * Copyright (c) 2010-2014 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <dzo@simtreas.ru>
 *
 * See REAME (rus)
 */

#include "libbb.h"
#include <dirent.h>
#include <sys/wait.h>
#include <utime.h>
#include "arch_crc.h"
#include "km.h"

#define TICK 20         /* sec */
#define HIDENAME ".hide.tmp"


struct FLIST {
	char *fullname;    /* initial absolute file name */
	char *n_localname; /* encoded fullname without work_dir and may be added [number_work_dir] */
	char *localname;   /* encoded fullname without work_dir */
	char *outname;
	char *archname;
	struct FLIST *l;
	off_t size;
	dev_t dev;      /* stat st_dev field unique type of file system */
	ino_t ino;      /* stat st_ino field unique in a file system */
	time_t mtime;   /* last modification */
	int status;     /* see above */
};

static struct FLIST *flist;

static size_t wd_len;           /* current work dir len + 1 */
static size_t od_len;           /* strlen(out_dir) + 1 */
static char *send_to;           /* out dir */
static int num_wd;              /* current workdir's number, zero for wihout set n_localname as [N]localname */


struct CHG_PATHS {
	char *from_name;
	char *to_name;
	struct CHG_PATHS *link;
};
static struct CHG_PATHS *chg_paths;

/* -A flags */
enum ARCH_MODES {
	arch_need_local = 1,            /* hardlink to source (but default mode - link to out file) */
	arch_need_copy = 2,             /* copy to archive, do not make hardlinks */
	arch_need_patch_change = 4,     /* converting paths before make a archive file name */
	arch_need_compare = 8,          /* find from archive */
	arch_need_uniq = 16,            /* find from archive and send to out uniq files only */
	arch_create_and_exit = 32,      /* exit after create the archive CRC data base */
	arch_need_uid_change = 64,
	arch_modes_delim = 128,
	arch_create_db = 256,
	arch_debug_list_arch = 512
};
static int arch_mode;
static int arch_modes[256] = {
	['l'] = arch_need_local,
	['c'] = arch_need_copy,
	['h'] = arch_need_patch_change,
	['C'] = arch_need_compare,
	['D'] = arch_need_uniq,
	['Q'] = arch_create_and_exit,
	['u'] = arch_need_uid_change,
	[':'] = arch_modes_delim
};

/* -S flags */
static int remove_subdirs, skip_first_dirs, no_subdirs;

static uid_t usage_euid, cur_euid;      /* must initialized by main() */
static uid_t out_euid;
static uid_t uid_for_delete;
static uid_t uid_for_archive;

static void seteuid_one_warn(uid_t uid)
{
	/* suppress identical syslog messages "can not set euid 'uid'" */
#define MAX_TST_SETEUID 4
	static uid_t is_not_ok_for_change[MAX_TST_SETEUID];
	static int max_initialized;

	if(uid == cur_euid)
		return;
	if(seteuid(uid)) {
		int i;

		for(i = 0; i < max_initialized; i++)
			if(is_not_ok_for_change[i] == uid)
				return;
		bb_perror_msg("can not set euid %d", uid);
		if(max_initialized == MAX_TST_SETEUID)
			i = 0;
		else
			max_initialized++;
		is_not_ok_for_change[i] = uid;
	} else {
		cur_euid = uid;
	}
}

static char *paths_change(const char *pfile)
{
    char *s = NULL;

    if(skip_first_dirs) {
	int n;

	for(n = 0; n < skip_first_dirs; n++) {
		s = strchr(pfile, '/');
		if(s == NULL)
			break;
		pfile = s + 1;
	}
    } else {
	s = strrchr(pfile, '/');
    }
    if(s == NULL || remove_subdirs) {
	s = concat_path_file(send_to, s == NULL ? pfile : s);
    } else {
	char *file = bb_xstrdup(pfile);
	char *sub;
	struct CHG_PATHS *p;
	size_t l;

	if(skip_first_dirs) {
		s = strrchr(pfile, '/');
		if(s == NULL)   /* have not subdir more */
			return concat_path_file(send_to, pfile);
	}
	s = file + (s - pfile);
	/* filename have path, convert paths if need */
	*s = '\0';      /* without filename */
	for(p = chg_paths; p != NULL; p = p->link) {
		sub = strstr(file, p->from_name);
		if(sub == NULL)
			continue;
		l = strlen(p->from_name);
		/* test really subpath, it is ok if
		   ^sub$ or ^sub/ or /sub$ or /sub/ */
		if((sub == file || sub[-1] == '/') &&
			  (sub[l] == '/' || sub[l] == '\0')) {
			/* change */
			char *e;

			if(sub == file)
				e = bb_xasprintf("%s%s/%s", p->to_name, sub+l, s+1);
			else {
				sub[-1] = '\0';
				e = bb_xasprintf("%s/%s%s/%s", file, p->to_name, sub+l, s+1);
			}
			free(file);
			file = e;
			s = strrchr(e, '/');
			*s = '\0';
		}
	}
	*s = '/';
	s = concat_path_file(send_to, file);
	free(file);
    }
	return s;
}



/* from my refer.c */
static const unsigned char utf8_koi8r_d0[] =
	"";
static const unsigned char utf8_koi8r_d1[] =
	"";

static void UTF8_2_koi8r(char *str)
{
	unsigned char *s = (unsigned char *)str;
	unsigned char *o = s;

	while(*s != '\0') {
		if(*s == 208 && *(s+1) >= 144 && *(s+1) <= 191) {
			s++;
			*o++ = utf8_koi8r_d0[*s - 144];
		} else if(*s == 209 && *(s+1) >= 128 && *(s+1) <= 143) {
			s++;
			*o++ = utf8_koi8r_d1[*s - 128];
		} else if(*s == 208 && *(s+1) == 129) {
			s++;
			*o++ = '';
		} else if(*s == 209 && *(s+1) == 145) {
			s++;
			*o++ = '';
		} else if(*s == 226 && *(s+1) == 132 && *(s+2) == 150) {
			s += 2;
			*o++ = 'N';
		} else {
			*o++ = *s;
		}
		s++;
	}
	*o = 0;
}

static char *koi8r_2_UTF8(char *str)
{
	char *s, *d;
	int need = 0;
	char *out, *out0;

	for(s = str; *s != '\0'; s++) {
		int c = U_C(*s);

		if(strchr((char *)utf8_koi8r_d0, c) != NULL || strchr((char *)utf8_koi8r_d1, c) ||
				c == U_C('') || c == U_C(''))
			need++;
	}
	if(!need)
		return str;
	out0 = out = xmalloc((s-str)+need+1);
	for(s = str; *s != '\0'; s++) {
		int c = U_C(*s);

		if((d = strchr((char *)utf8_koi8r_d0, c)) != NULL) {
			*out++ = (char)0xd0;
			c = 144+(d-(char *)utf8_koi8r_d0);
		} else if((d = strchr((char *)utf8_koi8r_d1, c)) != NULL) {
			*out++ = (char)0xd1;
			c = 128+(d-(char *)utf8_koi8r_d1);
		} else if(c == U_C('')) {
			*out++ = (char)0xd0;
			c = 129;
		} else if(c == U_C('')) {
			*out++ = (char)0xd1;
			c = 145;
		}
		*out++ = c;
	}
	*out = '\0';
	free(str);
	return out0;
}

enum ENCODES {
	need_encode_chr_first = 1,
	need_UTF8_2_koi8r = 2, need_koi8r_2_UTF8 = 4,
	need_koi8r_2_win = 8, need_win_2_koi8r = 16,
	need_encode_chr_second = 32
};
static int need_encode_local, need_encode_out, need_set_num_wd;

static char encode_to_local[512];
static char encode_to_out[512];

static void encode(char *s, const char *tbl)
{
	while(*s != '\0') {
		*s = tbl[U_C(*s)];
		s++;
	}
}


static char koi8_win_tbl[256];
static char win_koi8_tbl[256];

static char *encode_all(char *s, int encode_flags, char *tbl)
{
	if((encode_flags & need_encode_chr_first))
		encode(s, tbl);
	if((encode_flags & need_UTF8_2_koi8r))
		UTF8_2_koi8r(s);
	if((encode_flags & need_koi8r_2_win))
		encode(s, koi8_win_tbl);
	if((encode_flags & need_win_2_koi8r))
		encode(s, win_koi8_tbl);
	if((encode_flags & need_koi8r_2_UTF8))
		s = koi8r_2_UTF8(s);
	if((encode_flags & need_encode_chr_second))
		encode(s, tbl+256);
	return s;
}



static inline int match_mask__or_null_mask(const char *mask, const char *tst)
{
	return mask == NULL || match(mask, tst) != FNM_NOMATCH;
}

static char *in_mask;

static void add_file(char *fullname, struct stat *pst)
{
	struct FLIST *fl;
	char *localname;

	for(fl = flist; fl != NULL; fl = fl->l) {
		if(fl->dev == pst->st_dev && fl->ino == pst->st_ino) {
			/* have in list already */
			fl->status &= ~STATUS_TSTD;
			if((fl->status & STATUS_BADO)) {
				if(access(fullname, R_OK) == 0)
					fl->status &= ~STATUS_BADO;
			}
			if(fl->mtime != pst->st_mtime) {
				fl->mtime = pst->st_mtime;
				fl->status = STATUS_SIZE;
			}
			break;
		}
	}
	if(fl != NULL) { /* is not new */
		free(fullname);
		return;
	}

	localname = fullname + wd_len;
	if(need_encode_local)
		localname = encode_all(bb_xstrdup(localname), need_encode_local, encode_to_local);

	if(!match_mask__or_null_mask(in_mask, localname)) {
		if(!daemon_mode)
			bb_error_msg("skip '%s' - is not match with '%s'", localname, in_mask);
		if(need_encode_local)
			free(localname);
		free(fullname);
		return;
	}

	fl = xmalloc(sizeof(struct FLIST));
	fl->fullname = fullname;
	fl->localname = localname;
	if(num_wd)
		fl->n_localname = bb_xasprintf("[%d]%s", num_wd, localname);
	else
		fl->n_localname = localname;

	if(!daemon_mode)
		bb_error_msg("added %s", fl->n_localname);

	fl->outname = paths_change(localname);
	if((arch_mode & arch_need_patch_change) != 0)
		fl->archname = bb_xstrdup(fl->outname + strlen(send_to) + 1);
	else
		fl->archname = localname;

	fl->outname = encode_all(fl->outname, need_encode_out, encode_to_out);

	fl->l = flist;
	flist = fl;
	fl->size = pst->st_size;
	fl->dev = pst->st_dev;
	fl->ino = pst->st_ino;
	fl->mtime = pst->st_mtime;
	fl->status = STATUS_SIZE;
}


typedef struct llist_s {
	char *data;
	struct llist_s *link;
} llist_t;


/* from my mkdep */
/* emulate recursion. One open a directory in each time,
   memory used only for save subdirs names in the current opened directory.
   Result is not good for see - it have strange sorting, but scan is full,
   fast and minimal use resourses */
static void recursive_scan_dir(const char *wd)
{
    llist_t dirs0;
    llist_t *dirs = &dirs0;

    dirs->data = (char *)wd;
    do {
	llist_t *d_add = NULL;
	do {
	    DIR *dir = opendir(dirs->data);

	    if (dir == NULL) {
		if(daemon_mode)
			bb_perror_msg_and_die("opendir(%s)", dirs->data);
		else
			bb_perror_msg("opendir(%s)", dirs->data);
	    } else {
		struct dirent *de;

		while ((de = readdir(dir)) != NULL) {
			/* parse_entry */
			if(de->d_name[0] != '.') {
				struct stat st;
				char *fp = concat_path_file(dirs->data, de->d_name);

				if(lstat(fp, &st)) {
					if(match_mask__or_null_mask(in_mask, de->d_name))
						bb_perror_msg("stat(%s)", fp);
				} else {
					if(S_ISREG(st.st_mode)) {
						add_file(fp, &st);
						continue;
					} else if(S_ISDIR(st.st_mode) && no_subdirs == 0) {
						/* direntry is directory. */
						llist_t *new_head = xmalloc(sizeof(llist_t));

						new_head->data = fp;
						new_head->link = d_add;
						d_add = new_head;
						continue;
					}
				}
				/* is not file or dir */
				free(fp);
			}
		}
		closedir(dir);
	    }
	    if(dirs != &dirs0) {
		llist_t *tmp = dirs;

		free(dirs->data);
		dirs = dirs->link;
		free(tmp);
	    } else {
		break;
	    }
	} while(dirs);
	dirs = d_add;   /* next subdirs level */
    } while(dirs);
}


static char *hidename;
static int nobinary;
static int nodelete;
static int no_mkdir;
static char *archive_mask;
static char *overwrite_mask;


static void move(struct FLIST *fl)
{
	FILE *in;
	char *outname = fl->outname;
	sigset_t m;     /* a signal mask for block in critical section */
	int status = 0;
	unsigned int crc;
	char *s;
	int need_unblock = 0;

	if(access(outname, F_OK) == 0) {
		if(fl->status != STATUS_ALREADY) {
			if(overwrite_mask == NULL ||
				match(overwrite_mask, fl->localname) == FNM_NOMATCH) {
					fl->status = STATUS_ALREADY;
					if((arch_mode & (arch_need_compare|arch_need_uniq))
							!= (arch_need_compare|arch_need_uniq)) {
						log_msg("%s already present", fl->n_localname);
						return;
					}
					/* "-A AC" mode: compare now, may be eq and delete it */
			} else {
				log_msg("%s already present, rewroted by mask", fl->n_localname);
			}
		} else {
			return;
		}
	}

	/* initial privilege */
	in = fopen(fl->fullname, "r");
	if(in == NULL) {
		bb_perror_msg("can not open %s for read", fl->n_localname);
		fl->status |= STATUS_BADO;
		return;
	}

	/* user privilege for create file */
	seteuid_one_warn(out_euid);

	status = read_file_crc_tst0_out(fl->n_localname, in, &crc, nobinary, &(fl->size), hidename, "w\0temporary");
	if(status)
		goto err_out;

	if(!no_mkdir) {
	  /* Auto-create directories. */
	  for(s = outname + od_len; *s; s++) {
		if (*s == '/') {
			*s = '\0';
			if (mkdir(outname, 0755) != 0 && errno != EEXIST) {
				bb_perror_msg("mkdir(%s)", outname);
				status = STATUS_IOE;
				*s = '/';
				goto err_out;
			}
			*s = '/';
		}
	  }
	}

	/* block signals before enter critical section */
	m = bb_sigblock((1<<SIGTERM)|(1<<SIGQUIT)|(1<<SIGINT)|(1<<SIGHUP));
	need_unblock = 1;

	if(archive != NULL && match_mask__or_null_mask(archive_mask, fl->localname)) {
		char *s;

		seteuid_one_warn(uid_for_archive);

		if((arch_mode & (arch_need_compare|arch_need_uniq)) != 0
				&& fl->size != 0 &&
				(status = find_from_db(in, fl->size, &s, crc)) == STATUS_DUPA) {
			char *new_symlink;

			log_msg("%s=%s", fl->n_localname, s);
			new_symlink = make_archive_name(fl->archname);
			symlink(s, new_symlink);
			free(new_symlink);
			if((arch_mode & arch_need_uniq) == 0)
				status = STATUS_DUP_TOO;
		}

		if(fl->status == STATUS_ALREADY && status != STATUS_DUPA) {
			/* not eq, return without other changes */
			status = STATUS_ALREADY;
			log_msg("%s already present", fl->n_localname);
		}

		if(status == STATUS_IOE) {
			bb_perror_msg("input file problem %s", fl->n_localname);
		} else if(status == 0 || status == NEW_UNIQ) {
			int need_crc_store;

			if(status == NEW_UNIQ) {
				status = 0;
				need_crc_store = 1;
			} else {
				need_crc_store = 0;
			}
			s = make_archive_name(fl->archname);
			if((arch_mode & (arch_need_local|arch_need_copy)) == 0) {
				/* default archive mode - hadrlink to hidename file */
				if(link(hidename, s))
					bb_perror_msg_and_die("link %s to %s", hidename, s);
			} else if((arch_mode & arch_need_local) != 0) {
				if(link(fl->fullname, s))
					bb_perror_msg_and_die("link %s to %s", fl->fullname, s);
			} else {
				/* coping */
				rewind(in);
				status = read_file_crc_tst0_out(fl->n_localname, in, NULL, 0, &(fl->size), s, "w\0archive");
				if(status)
					unlink(s);
			}
			if(need_crc_store && status == 0)
				store_crc_file(s, fl->size, crc);
			free(s);
		}
		seteuid_one_warn(out_euid);
	}

	/* rename hidden to out */
	if(status == 0 || status == STATUS_DUP_TOO) {
		status = rename(hidename, outname);
		if(status) {
			bb_perror_msg("rename %s to %s", hidename, outname);
			status = STATUS_IOE;
		} else if(!daemon_mode)
			bb_error_msg("%s to %s renamed", hidename, outname);
	}

	if(!status)
		log_msg("%s (%lld)", fl->n_localname, (long long)fl->size);

err_out:
	if(status)
		unlink(hidename);

	if(status == 0 || status == STATUS_DUPA) {
		if(!nodelete) {
			seteuid_one_warn(uid_for_delete);
			if(unlink(fl->fullname))
				bb_perror_msg_and_die("unlink %s", fl->n_localname);
		} else {
			bb_error_msg("skip unlink %s", fl->n_localname);
		}
		status = STATUS_DONE;
	}

	if(need_unblock)
		bb_sigunblock(m);
	/* end critical section */
	fclose(in);

	/* restore initial privilegies */
	seteuid_one_warn(usage_euid);

	fl->status = status;
}


static volatile sig_atomic_t interrupted;

static void sighup_handler(int sig)
{
	if(sig == 0) {
		/* set mode */
		struct sigaction act;

		memset(&act, 0, sizeof(act));
		act.sa_handler = sighup_handler;
		act.sa_flags |= SA_RESTART;
		sigaddset(&act.sa_mask, SIGHUP);
		while(sigaction(SIGHUP, &act, NULL) < 0 && errno == EINTR)
			;
	} else {
		/* interrupt mode */
		interrupted = 1;
	}
}

static char *chk_path_pair(char *p, const char *name, const char *pcp, int line)
{
	/* "/xxx/" convert to "xxx" */
	char *s;

	while(*p == '/')
		p++;
	if(*p == '\0')
		bb_error_msg_and_die("%s:%d the %s path can not set as /", pcp, line, name);
	s = p + strlen(p) - 1;
	while(s > p && *s == '/') {
		*s = '\0';
		s--;
	}
	return p;
}

static void add_path_pair(char *from, char *to, const char *pcp, int line)
{
	struct CHG_PATHS *p = xmalloc(sizeof(struct CHG_PATHS));

	p->from_name = chk_path_pair(from, "from", pcp, line-2);
	p->to_name = chk_path_pair(to, "to", pcp, line-1);
	p->link = chg_paths;
	chg_paths = p;
}

static void load_convert_paths(const char *pcp)
{
	FILE *in;
	char *s;
	int line;
	char *from, *to;

	if(pcp == NULL)
		return;

	in = xwafopen(pcp, "r\0");

	from = to = NULL;
	line = 0;
	while(fgets(bufc, sizeof(bufc), in) != NULL) {
		line++;
		s = strchr(bufc, '\n');
		if(s == NULL)
			bb_error_msg_and_die("%s:%d line too long", pcp, line);
		*s = '\0';
		s = strchr(bufc, '\r');
		if(s != NULL)
			*s = '\0';
		s = bufc;
		if(*s == '#' || *s == '\0') {
			if(from == NULL && to == NULL)
				continue;
			if(from != NULL && to != NULL) {
				add_path_pair(from, to, pcp, line);
				from = to = NULL;
				continue;
			}
			bb_error_msg_and_die("%s:%d can not find pair.\nNeed from_line to_line empty_or_#lines", pcp, line);
		}
		s = bb_xstrdup(s);
		if(from == NULL)
			from = s;
		else if(to == NULL)
			to = s;
		else
			bb_error_msg_and_die("%s:%d can not parse pair.\nNeed from_line to_line empty_or_#lines", pcp, line);
	}
	fclose(in);
	if(from != NULL) {
		if(to != NULL)
			add_path_pair(from, to, pcp, line);
		else
			bb_error_msg_and_die("%s:%d can not find pair.\nNeed from_line to_line empty_or_#lines", pcp, line);
	}
}

static void scan_proc(void)
{
	DIR *dir_proc, *dir_fd;
	struct dirent *entry;
	static char buf_fd[272];        /* /proc/[0-9]{255}/fd */
	static char buf_file[530];      /* /proc/[0-9]{255}/fd/[0-9]{255}/ */
	struct stat st;
	struct FLIST *fl;

	/* drop previous set STATUS_PROC */
	for(fl = flist; fl != NULL; fl = fl->l)
		fl->status &= ~STATUS_PROC;

	dir_proc = opendir("/proc");

	if (dir_proc == NULL)
		bb_perror_msg_and_die("/proc");
	while ((entry = readdir(dir_proc)) != NULL) {
		/* search /proc/pid (numeric only) */
		char *e = entry->d_name;
		while(*e >= '0' && *e <= '9')
			e++;
		if(*e != '\0' || e == entry->d_name)
			continue;

		snprintf(buf_fd, sizeof(buf_fd), "/proc/%s/%s", entry->d_name, "fd");
		dir_fd = opendir(buf_fd);
		if (dir_fd == NULL)        /* pid is quit? */
			continue;

		while ((entry = readdir(dir_fd)) != NULL) {
			if(entry->d_name[0] == '.')
				continue;

			snprintf(buf_file, sizeof(buf_file), "%s/%s", buf_fd, entry->d_name);
			if(stat(buf_file, &st))
				continue;

			for(fl = flist; fl != NULL; fl = fl->l) {
				if(fl->dev == st.st_dev && fl->ino == st.st_ino) {
					/* drop STATUS_IOE too */
					if(fl->status != STATUS_DONE) {
						fl->status = STATUS_PROC;
						if(!daemon_mode)
							bb_error_msg("/proc found %s", fl->fullname);
					}
					break;
				}
			}
		}
		closedir(dir_fd);
	}
	closedir(dir_proc);

	/* if do not set a work status then set STATUS_SIZE */
	for(fl = flist; fl != NULL; fl = fl->l)
		if(fl->status == 0)
			fl->status = STATUS_SIZE;
}

static const char *priority_sort;
static void sort(int flg_sort_mtime, int flg_rev)
{
	struct FLIST *fl1, *fl2, fl_tmp, *l;

	for(fl1 = flist; fl1 != NULL; fl1 = fl1->l)
		for(fl2 = fl1->l; fl2 != NULL; fl2 = fl2->l) {
			int s1, s2;

			if(priority_sort != NULL) {
				int pm1 = match(priority_sort, fl1->localname) != FNM_NOMATCH;
				int pm2 = match(priority_sort, fl2->localname) != FNM_NOMATCH;
				s1 = pm1 - pm2;
			} else {
				s1 = 0;
			}
			if(flg_sort_mtime)
				s2 = flg_rev ? fl1->mtime < fl2->mtime : fl1->mtime > fl2->mtime;
			else
				s2 = 0;

			if(s1 < 0 || (s1 == 0 && s2)) {
				memcpy(&fl_tmp, fl1, sizeof(fl_tmp));
				fl_tmp.l = fl2->l;
				l = fl1->l;
				memcpy(fl1, fl2, sizeof(fl_tmp));
				fl1->l = l;
				memcpy(fl2, &fl_tmp, sizeof(fl_tmp));
			}
		}
}

static long set_time_opt_form_str(const char *s, long def)
{
	if(s != NULL) {
		int t = safe_strtol(s);

		if(errno != 0 || t <= 0)
			bb_error_msg_and_die("strange time value %s", s);
		return t;
	}
	return def;
}


static char *setproctitle_a0, *setproctitle_LastArg;

static void prepare_setproctitle(char **argv)
{
	char **av0 = argv;

	setproctitle_a0 = argv[0];
	while (*argv != NULL)
		argv++;
	setproctitle_LastArg = argv[-1] + strlen (argv[-1]);
	daemon_name = *av0;     /* bb_xstrdup() may be produce a error */
	while (*av0 != NULL) {
		*av0 = bb_xstrdup(*av0);
		av0++;
	}
}

static void setproctitle(const char *format, ...)
{
	va_list parms;

	va_start(parms, format);
	setproctitle_a0 += vsnprintf(setproctitle_a0,
			setproctitle_LastArg - setproctitle_a0, format, parms) + 1;
	while (setproctitle_a0 < setproctitle_LastArg)
		*setproctitle_a0++ = ' ';
	va_end(parms);
}

static volatile int alarmed;
static volatile int heart_beat;
static volatile int terminate;
static void handler(int s)
{
	switch(s) {
		case SIGALRM:
			alarmed = s;
			break;
		case SIGUSR1:
			heart_beat = s;
			break;
		case SIGHUP:
		case SIGTERM:
			terminate = s;
			break;
	}
}

static void set_signal_intr(int signo, sighandler_t h)
{
	struct sigaction act;

	memset(&act, 0, sizeof(act));
	act.sa_handler = h;
	while(sigaction(signo, &act, NULL) < 0 && errno == EINTR)
		;
}

static void run_script(char *script, char *opt)
{
	int k_pid;
	volatile int vfork_exec_errno;
	char *kargv[3];

	kargv[0] = (char *)script;
	kargv[1] = opt;
	kargv[2] = NULL;
	log_msg("start %s %s", script, opt);

	vfork_exec_errno = 0;
	k_pid = vfork();

	if(k_pid < 0)
		bb_perror_msg_and_die("vfork() for kill");
	if(k_pid == 0) {
		execvp(script, kargv);
		vfork_exec_errno = errno;
		_exit(-1);
	}
	if(vfork_exec_errno) {
		errno = vfork_exec_errno;
		bb_perror_msg_and_die("can not start %s %s", script, opt);
	}
	while(waitpid(k_pid, NULL, 0) != k_pid)
		;
}

static int restart0_time;
static int restart_time;
static int heart_time;

static void run(char **argv, char *script, int foreground)
{
	int pid, c_pid, status;
	volatile int vfork_exec_errno;

	set_signal_intr (SIGALRM, handler);
	set_signal_intr (SIGUSR1, handler);
	set_signal_intr (SIGTERM, handler);
	set_signal_intr (SIGHUP,  handler);

	for(;;) {
		vfork_exec_errno = 0;
		pid = vfork();
		if(pid < 0)
			bb_perror_msg_and_die("vfork()");
		if(pid == 0) {
			execvp(argv[0], argv);
			vfork_exec_errno = errno;
			_exit(-1);
		}
		if(vfork_exec_errno) {
			errno = vfork_exec_errno;
			bb_perror_msg_and_die("can not start %s", argv[0]);
		}
		log_msg("%s started, pid=%d", daemon_name, pid);
		for(c_pid = 0; c_pid != pid; ) {
			if(foreground)
				log_msg("heart_beat=%d terminate=%d", heart_beat, terminate);
			alarm(heart_time);
			c_pid = heart_beat == 0 && terminate <= 0 ? wait(&status) : 0;
			if(c_pid <= 0) {
				if(heart_beat) {
					if(foreground) {
						printf("%s heart beat", daemon_name);
						if(alarmed)
							printf(" and alarmed");
						printf("\n");
					}
					heart_beat = 0;
				}
				if(alarmed) {
					if(terminate < 0) {
						log_msg("%s alarmed while wait kill daemon", daemon_name);
						if(script != NULL)
							run_script(script, "-k");
						 else
							kill(pid, SIGKILL);
					} else {
						log_msg("%s alarmed daemon", daemon_name);
						terminate = SIGALRM;
					}
				}
				if(terminate > 0) {
					if(terminate == SIGTERM) {
						log_msg("kill SIGTERM %s", daemon_name);
						kill(pid, SIGTERM);
						terminate = -SIGTERM;
					} else if(terminate == SIGHUP) {
						log_msg("kill SIGKILL %s", daemon_name);
						kill(pid, SIGKILL);
						terminate = -SIGTERM;
					} else {
						kill(pid, SIGTERM);
						terminate = -SIGALRM;
					}
				}
			}
			alarm(0);
			alarmed = 0;
		}
		if (WIFEXITED(status))
			log_msg("%s has exited with status=%d", daemon_name, WEXITSTATUS(status));
		if (WIFSIGNALED(status))
			log_msg("%s has exited with signal=%d", daemon_name, WTERMSIG(status));
		if(terminate == -SIGTERM || terminate > 0)
			return;
		status = WIFEXITED(status) && WEXITSTATUS(status) == 0 ? restart0_time : restart_time;
		if(foreground)
			log_msg("sleep %d", status);
		sleep(status);
		if(terminate > 0)        /* while sleep */
			return;
		alarmed = 0;
		heart_beat = 0;
		terminate = 0;
	}
}

static const unsigned char koi8_win[] = {
/* 128  129  130  131  132  133  134  135 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 136  137  138  139  140  141  142  143 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 144  145  146  147  148  149  150  151 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 152  153  154  155  156  157  158  159 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 160  161  162  163  164  165  166  167 */
     0,   0,   0,'\270',0,   0,   0,   0,
/* 168  169  170  171  172  173  174  175 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 176  177  178  179  180  181  182  183 */
     0,   0,   0,'\250',0,   0,   0,   0,
/* 184  185  186  187  188  189  190  191 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 192  193  194  195  196  197  198  199 */
   '', '', '', '', '', '', '', '',
/* 200  201  202  203  204  205  206  207 */
   '', '', '', '', '', '', '', '',
/* 208  209  210  211  212  213  214  215 */
   '', '', '', '', '', '', '', '',
/* 216  217  218  219  220  221  222  223 */
   '', '', '', '', '', '', '', '',
/* 224  225  226  227  228  229  230  231 */
   '', '', '', '', '', '', '', '',
/* 232  233  234  235  236  237  238  239 */
   '', '', '', '', '', '', '', '',
/* 240  241  242  243  244  245  246  247 */
   '', '', '', '', '', '', '', '',
/* 248  249  250  251  252  253  254  255 */
   '', '', '', '', '', '', '', '',
};

static const unsigned char win_koi8[] = {
/* 128  129  130  131  132  133  134  135 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 136  137  138  139  140  141  142  143 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 144  145  146  147  148  149  150  151 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 152  153  154  155  156  157  158  159 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 160  161  162  163  164  165  166  167 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 168  169  170  171  172  173  174  175 */
   '',   0,   0,   0,   0,   0,   0,   0,
/* 176  177  178  179  180  181  182  183 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 184  185  186  187  188  189  190  191 */
   '',   0,   0,   0,   0,   0,   0,   0,
/* 192  193  194  195  196  197  198  199 */
   '', '', '', '', '', '', '', '',
/* 200  201  202  203  204  205  206  207 */
   '', '', '', '', '', '', '', '',
/* 208  209  210  211  212  213  214  215 */
   '', '', '', '', '', '', '', '',
/* 216  217  218  219  220  221  222  223 */
   '', '', '', '', '', '', '', '',
/* 224  225  226  227  228  229  230  231 */
   '', '', '', '', '', '', '', '',
/* 232  233  234  235  236  237  238  239 */
   '', '', '', '', '', '', '', '',
/* 240  241  242  243  244  245  246  247 */
   '', '', '', '', '', '', '', '',
/* 248  249  250  251  252  253  254  255 */
   '', '', '', '', '', '', '', ''
};


static void set_encodes(const char *t)
{
	int i;
	int *pneed_encode = &need_encode_local;
	char *encode_to = encode_to_local;

	if(t == NULL)
		return;
	for(i = 0; i < 160; i++)
		koi8_win_tbl[i] = win_koi8_tbl[i] = i;
	for(i = 160; i < 256; i++) {
		if(koi8_win[i-128])
			koi8_win_tbl[i] = koi8_win[i-128];
		if(win_koi8[i-128])
			win_koi8_tbl[i] = win_koi8[i-128];
	}
	for(i = 0; i < 256; i++)
		encode_to_local[i] = encode_to_local[i+256] =
			encode_to_out[i] = encode_to_out[i+256] = i;

	while(1) {
		/* ckeck format, do not accept empty "", "*o" and "*o:" */
		int c2;

		switch(*t) {
			case 'o':
				if(pneed_encode == &need_encode_out) {
err_es:
					bb_error_msg_and_die("invalid format -E encode_str");
				}
				if(*++t == ':')
					t++;
				pneed_encode = &need_encode_out;
				encode_to = encode_to_out;
				continue;       /* next if(*t == '\0') is error */
			case '8':
				*pneed_encode |= need_UTF8_2_koi8r;
				break;
			case '9':
				*pneed_encode |= need_koi8r_2_UTF8;
				break;
			case 'w':
				*pneed_encode |= need_koi8r_2_win;
				break;
			case 'W':
				*pneed_encode |= need_win_2_koi8r;
				break;
			case 'n':
				need_set_num_wd = 1;    /* always [1] */
				break;
			case 'N':
				need_set_num_wd = -1;   /* always without [N] */
				break;
			case '\'':
				/* check '?'-'?' */
				i = *++t;
				if(i == 0 || *++t != '\'')
					goto err_es;
				if(*++t != '-')
					goto err_es;
				t++;
				if(*t >= '0' && *t <= '3')
					goto oct2_enc;  /* '?'-OOO */
				if(*t != '\'')
					goto err_es;

			char2_enc:
				if(*++t == '\0')
					goto err_es;
				c2 = *t++;
				if(*t != '\'')
					goto err_es;
				goto set_c2;
			case '0': case '1': case '2': case '3':
				/* check OOO-OOO */
				if(strlen(t) < 7)
					goto err_es;
				if(!(t[1] >= '0' && t[1] <= '7'))
					goto err_es;
				if(!(t[2] >= '0' && t[2] <= '7'))
					goto err_es;
				if(t[3] != '-')
					goto err_es;
				i = ((int)(t[0]-'0'))*64 + ((int)(t[1]-'0'))*8 + ((int)(t[2]-'0'));
				t += 4;
				if(*t == '\'')
					goto char2_enc; /* OOO-'?' */

				if(!(t[0] >= '0' && t[0] <= '3'))
					goto err_es;
			oct2_enc:
				if(!(t[1] >= '0' && t[1] <= '7'))
					goto err_es;
				if(!(t[2] >= '0' && t[2] <= '7'))
					goto err_es;
				c2 = ((int)(t[0]-'0'))*64 + ((int)(t[1]-'0'))*8 + ((int)(t[2]-'0'));
				t += 2;

			set_c2:
				if(*pneed_encode == 0) {
					*pneed_encode = need_encode_chr_first;
					encode_to[i] = c2;
				} else if(*pneed_encode == need_encode_chr_first) {
					encode_to[i] = c2;
				} else {
					*pneed_encode |= need_encode_chr_second;
					encode_to[i+256] = c2;
				}
				break;
			default:
				bb_error_msg_and_die("invalid format -E encode_str");
		}
		if(*++t == ':')
			t++;
		if(*t == '\0')
			return;
	}
}

static int set_arch_modes(char *archive_mode)
{
	int mode;

	uid_for_archive = usage_euid;

	while(archive_mode != NULL && *archive_mode != '\0') {
		if(archive == NULL)
			bb_error_msg_and_die("The '-A archive_mode' option need the '-a archive_dir' option");
		if((mode = arch_modes[U_C(*archive_mode)]) == 0)
			bb_error_msg_and_die("Unsupport -A %c mode", *archive_mode);
		arch_mode |= mode;
		if(mode == arch_need_uid_change) {
			char *colon;

			if(*++archive_mode != '=')
				bb_error_msg_and_die("stange syntax -A, need %c=user_for_archiving", *archive_mode);
			if((colon = strchr(++archive_mode, ':')) != NULL)
				*colon = '\0';
			uid_for_archive = bb_xgetpwnam(archive_mode);
			if(colon != NULL)
				*colon++ = ':';
			archive_mode = colon;
		} else
			archive_mode++;
	}

	if((arch_mode & (arch_need_local|arch_need_copy)) == (arch_need_local|arch_need_copy))
		bb_error_msg_and_die("Conflict -A 'l' and 'c' modes");

	if(archive != NULL) {
		archive = make_fullpath_and_tstisdir(archive);
		if((arch_mode & arch_create_db) != 0) {
			seteuid_one_warn(uid_for_archive);
			fix_crc_file();
			if((arch_mode & arch_create_and_exit) != 0)
				return 1;       /* exit main */
			seteuid_one_warn(usage_euid);
			if((arch_mode & (arch_need_compare|arch_need_uniq)) == 0) {
				bb_error_msg("Warning: used '-A C' mode");
				arch_mode |= arch_need_compare;
			}
		} else if((arch_mode & (arch_need_compare|arch_need_uniq)) != 0) {
			if((arch_mode & arch_create_and_exit) != 0)
				bb_error_msg_and_die("Unsupport '-A Q' without '-c'");
			seteuid_one_warn(uid_for_archive);
			init_crc_list(arch_mode & arch_debug_list_arch);
			seteuid_one_warn(usage_euid);
		}
	} else if((arch_mode & arch_create_db) != 0)
		bb_error_msg_and_die("The -c option need the '-a archive_dir' option");
	return 0;
}

const char usage_msg[] =
"-s sendto_dir [-t sendto_temp_dir] [-i r0:rnot0[:hb]] [-0(no binary)]\n"
"[-f(oreground and debug)] [-p pidfile] [-S skip_first_dirs] [-1TI]\n"
"[-u uid_for_create] [-U uid_for_delete] [-e(mpty is ok too)] [-m mask]\n"
"[-R(do_not_create_subdirs)] [-l logfile] [-H hide_name] [-O overwrite_mask]\n"
"[-h file_with_convert_paths] [-n ticks] [-D(skip_delete)] [-2 rescan_time]\n"
"[-P(roc_tst)] [-E encode_str] [-C wait_need_remote[:script]]\n"
"[-x|X(sort_modify_time|reverse)] [-y priority_for_sort_mask] [-N daemon_name]\n"
"[-a archive [-M archive_mask] [-A archive_mode [-c]]] work_dir1 ...";

int main(int argc, char **argv)
{
	int flag, i;
	const char *need_encode = NULL;
	char *str_uid = NULL;
	char *str_uid_del = NULL;
	char *logfile = NULL;
	struct stat st;
	char *pcp = NULL;
	char *s_tick = NULL;
	long ticks;
	char *pidfile = NULL;
	const char **wds;        /* work dirs */
	size_t *wd_lens;         /* wd_lens[argc] = strlen(argv[argc]) + 1 */
	char *hn = HIDENAME;
	char *srescan_time = NULL;
	long rescan_time;
	char *s_skip_first_dirs = NULL;
	int  pid_parent = 0;

	int monitored = 0;
	char *s_monitor_time = NULL;
	char *kill_script = NULL;
	char *require_remote = NULL;
	long wait_restart_remote;
	char *script_restart_remote = NULL;
	char *archive_mode = NULL;

	prepare_setproctitle(argv);

/* positioning dependence "fe01DRPxXTIc" */
#define OPT_FOREGROUD   1
#define OPT_EMPTY_IS_OK 2
#define OPT_NO_BINARY   4
#define OPT_ONE_CALL    8
#define OPT_NO_DELETE   16
#define OPT_SKIP_MKDIR  32
#define OPT_SCAN_PROC   64
#define OPT_SORT        128
#define OPT_SORT_REV    256
#define OPT_UTIME_EL    512
#define OPT_NO_EXIT1    1024
#define OPT_CREATE_DB   2048
	flag = xgetopt(argc, argv, "fe01DRPxXTIcE:h:s:u:U:l:t:n:a:p:m:H:M:2:y:S:N:i:k:C:O:A:\0-s",
					&need_encode, &pcp, &send_to, &str_uid,
					&str_uid_del,
					&logfile, &hidename, &s_tick,
					&archive, &pidfile, &in_mask,
					&hn, &archive_mask, &srescan_time,
					&priority_sort, &s_skip_first_dirs,
					&daemon_name, &s_monitor_time, &kill_script, &require_remote,
					&overwrite_mask, &archive_mode);

	set_encodes(need_encode);
	nobinary = flag & OPT_NO_BINARY;

	if(s_skip_first_dirs != NULL) {
		skip_first_dirs = safe_strtol(s_skip_first_dirs);
		if(errno != 0 || skip_first_dirs < -1)
			bb_error_msg_and_die("Strange -Skip_first_dirs value '%s'", s_skip_first_dirs);
		if(skip_first_dirs == -1) {
			remove_subdirs = skip_first_dirs;
			skip_first_dirs = 0;
		} else if(skip_first_dirs == 0)
			no_subdirs = 1;
	}
	if(pcp != NULL) {
		if(remove_subdirs)
			bb_error_msg_and_die("Conflict options -S -1 and -h");
		if(no_subdirs)
			bb_error_msg_and_die("Conflict options -S 0 and -h");
		load_convert_paths(pcp);
	}

	usage_euid = cur_euid = geteuid();
	if(str_uid != NULL)
		out_euid = bb_xgetpwnam(str_uid);
	if(str_uid_del != NULL)
		uid_for_delete = bb_xgetpwnam(str_uid_del);
	else
		uid_for_delete = usage_euid;

	ticks = set_time_opt_form_str(s_tick, TICK);
	rescan_time = set_time_opt_form_str(srescan_time, -1);

	if((flag & OPT_SORT) && (flag & OPT_SORT_REV))
		bb_error_msg_and_die("need -x or -X are mutually exclusive");

	if(logfile != NULL)
		log_f = xwafopen(logfile, "a\0log");

	if(require_remote != NULL) {
		char *pcoma = strchr(require_remote, ':');
		if(pcoma != NULL) {
			*pcoma++ = '\0';
			if(*pcoma != '\0')
				script_restart_remote = pcoma;
		}
	}
	wait_restart_remote = set_time_opt_form_str(require_remote, -1);
	if(script_restart_remote != NULL)
		script_restart_remote[-1] = ':';        /* restore for fork daemon */

	if((flag & OPT_CREATE_DB) != 0)
		arch_mode |= arch_create_db;
	if((flag & OPT_FOREGROUD) != 0)
		arch_mode |= arch_debug_list_arch;
	if(set_arch_modes(archive_mode))
		return 0;       /* create archive CRC data base and exit */

	if(s_monitor_time != NULL) {
		char *am = s_monitor_time;
		char *pcoma = strchr(s_monitor_time, ':');

		if(pcoma == NULL && s_monitor_time[0] == 'm' && s_monitor_time[1] == '\0') {
			monitored = 1;
			goto mt_ok;
		}

		setproctitle("%s -i %s", daemon_name, am);

		if(pcoma != NULL)
			*pcoma = '\0';
		restart0_time = set_time_opt_form_str(s_monitor_time, ticks);
		if(pcoma != NULL) {
			s_monitor_time = pcoma + 1;
			pcoma = strchr(s_monitor_time, ':');
			if(pcoma != NULL)
				*pcoma++ = '\0';
			restart_time = set_time_opt_form_str(s_monitor_time, ticks);
		} else {
			restart_time = restart0_time;
		}
		if(pcoma != NULL && *pcoma == '\0')
			pcoma = NULL;
		heart_time = set_time_opt_form_str(pcoma, ticks * 2);

		/* set the exec argument as MONITORED */
		strcpy(am, "m");
		if((flag & OPT_FOREGROUD) == 0) {
			xdaemon(1, pidfile);
			if(logfile != NULL)
				freopen(logfile, "a", stderr);
		}
		run(argv, kill_script, (flag & OPT_FOREGROUD));
		unlink_pidfile();
		return 0;

	} else if(kill_script != NULL)
		bb_error_msg_and_die("-k need monitoring option (-i) too");

mt_ok:

	/* end fast check. Next may be slow for example network problem with a remote file system */

	argc -= optind;
	argv += optind;
	wds = xmalloc(sizeof(char *) * argc);
	wd_lens = xmalloc(sizeof(size_t) * argc);
	for(i = 0; i < argc; i++) {
		/* make fullpaths before chdir(/) */
		wds[i] = make_fullpath_and_tstisdir(argv[i]);
		wd_lens[i] = strlen(wds[i]) + 1;
	}
	send_to = make_fullpath_and_tstisdir(send_to);
	od_len = strlen(send_to) + 1;

	/*
		need_set_num_wd = -1 is always without [N]
		need_set_num_wd = 1 is always [1]
		need_set_num_wd = 0 is set to work_dirs > 1 for add [1..work_dirs]
	*/
	if(need_set_num_wd < 0) {
		need_set_num_wd = 0;
	} else if(argc > 1 && need_set_num_wd == 0) {
		need_set_num_wd = argc;
	}

	if(require_remote != NULL) {
		for(i = 0; i < argc; i++)
			if(tst_is_not_one_fs(send_to, wds[i]))
				break;
		if(i < argc) {
			sleep(wait_restart_remote);
			if(script_restart_remote)
				run_script(script_restart_remote, "-C");
			return 65;
		}
	}

	if(hidename == NULL) {
		hidename = send_to;
	} else {
		hidename = make_fullpath_and_tstisdir(hidename);
		tst_is_one_fs(send_to, hidename);
	}
	hidename = concat_path_file(hidename, hn);

	if(archive != NULL) {
		if((arch_mode & arch_need_local) == 0) {
			if((arch_mode & arch_need_copy) == 0)
				tst_is_one_fs(send_to, archive);
		} else {
			for(i = 0; i < argc; i++)
				tst_is_one_fs(wds[i], archive);
		}
	}

	if(monitored) {
		flag |= OPT_FOREGROUD | OPT_ONE_CALL;
		daemon_mode = 1;        /* but debug off */
		pid_parent = getppid();
	}

	if((flag & OPT_FOREGROUD) == 0)
		xdaemon(1, pidfile);
	setproctitle("%s", daemon_name);

	nodelete = flag & OPT_NO_DELETE;
	no_mkdir = flag & OPT_SKIP_MKDIR;

	sighup_handler(0);      /* set handler */

	/* main loop */
	for(;;) {
		struct FLIST *fl;
		struct FLIST *prev, *next;
		time_t trescan1 = 0, trescan2 = 0;

		if(pid_parent)
			kill(pid_parent, SIGUSR1);      /* heart beat */

		/* test deleted */
		for(fl = flist; fl != NULL; fl = fl->l)
			fl->status |= STATUS_TSTD;

		for(i = 0; i < argc; i++) {
			wd_len = wd_lens[i];
			if(need_set_num_wd > 0)
				num_wd = need_set_num_wd > 1 ? i + 1 : 1;
			recursive_scan_dir(wds[i]);
		}

		for(fl = flist; fl != NULL; fl = fl->l) {
			/* test for deleted */
			if((fl->status & STATUS_TSTD) != 0)
				fl->status = STATUS_DONE;
		}

		if(flist == NULL) {
			if(!daemon_mode)
				bb_error_msg("Empty file list");
			if(monitored) {
				if(log_f != NULL) {
					if((flag & OPT_UTIME_EL) != 0)
						utime(logfile, NULL);
					else
						log_msg("Empty file list");
				}
			}
			if((flag & OPT_ONE_CALL) && (flag & OPT_NO_EXIT1) == 0) {
				unlink_pidfile();
				return 0;
			}
		}
		if((flag & OPT_SCAN_PROC))
			scan_proc();

		if((flag & (OPT_SORT|OPT_SORT_REV)) || priority_sort != NULL)
			sort(flag & (OPT_SORT|OPT_SORT_REV), flag & OPT_SORT_REV);

		if(rescan_time > 0)
			time(&trescan1);

		for(fl = flist; fl != NULL; fl = fl->l) {
			if((fl->status & (STATUS_SIZE|STATUS_WAIT|STATUS_ALREADY)) != 0) {
				if((fl->status & STATUS_BADO))
					continue;
				if(stat(fl->fullname, &st)) {
					/* removed ? */
					fl->status = STATUS_DONE;
					continue;
				}
				if(st.st_size == 0 && !(flag & OPT_EMPTY_IS_OK)) {
					if((flag & OPT_ONE_CALL) && trescan2 == 0)
						fl->status = STATUS_DONE;
					continue;
				}
				if(fl->size == st.st_size) {
					if(fl->status == STATUS_SIZE) {
						/* sleep the tick for safe */
						fl->status = STATUS_WAIT;
						continue;
					}
					if(str_uid == NULL)
						out_euid = st.st_uid;
					move(fl);
					if(pid_parent)
						kill(pid_parent, SIGUSR1);      /* heart beat */
					if(rescan_time > 0) {
						time(&trescan2);
						if((trescan2 - trescan1) > rescan_time)
							break;  /* need rescan */
						trescan2 = 0;
					}
					continue;
				}
				/* size changed */
				/* drop other statuses */
				fl->status = STATUS_SIZE;
				if(!daemon_mode && fl->size != st.st_size)
					bb_error_msg("%s changed size previous=%lld current=%lld",
							fl->n_localname,
							(long long)fl->size,
							(long long)st.st_size);
				fl->size = st.st_size;
			}
		}

		/* free list */
		for(prev = fl = flist; fl != NULL; fl = next) {
			next = fl->l;
			if(fl->status == STATUS_DONE || (trescan2 > 0 && fl->status != STATUS_IOE)) {
				free(fl->fullname);
				free(fl->outname);
				if(fl->archname != fl->localname)
					free(fl->archname);
				if(need_encode_local)
					free(fl->localname);
				if(need_set_num_wd)
					free(fl->n_localname);
				if(fl == flist)
					flist = prev = next;
				else
					prev->l = next;
				free(fl);
			} else {
				if(fl != flist)
					prev = prev->l;
			}
		}

		if(interrupted) {
			interrupted = 0;
			for(fl = flist; fl != NULL; fl = fl->l) {
				if(fl->status == STATUS_IOE)
					fl->status = STATUS_WAIT;
			}
		}

		if((flag & OPT_ONE_CALL) && trescan2 == 0) {
			for(fl = flist; fl != NULL; fl = fl->l) {
				if((fl->status & (STATUS_IOE|STATUS_ALREADY|STATUS_BADO))==0)
					break;
			}
			if(fl == NULL) {
				if((flag & OPT_NO_EXIT1) == 0 || flist != NULL) {
					unlink_pidfile();
					return (flist != NULL);
				}
			}
		}

		/* if trescan2 > 0 then we work more of rescan_time */
		if(trescan2 == 0)
			sleep(ticks);

	}
}
